/*

             ************************************************
             ****            VEGA - Ener V1.1            ****
             ****           Energy calculation           ****
             **** Copyright 1996-98, Alessandro Pedretti ****
             ************************************************

This code calculates the energy interaction between two molecules (e.g. a ligand
and a macromolecule) using the CVFF force field (Copyright Biosym/MSI).

*/


#include <math.h>
#include <stdio.h>
#include <string.h>

#include "globdef.h"
#include "globvar.h"
#include "globstr.h"

#define  CVFF_PAR	"CVFF.par"
#define  CONV_COUL      332.0


/**** Calculate non-bond interations between two molecules ****/

BOOL CalcEnerCVFF(FILE *OUT, ATOMO *InizAtm, ULONG TotAtm,
                  LONG NomeRes, LONG NRes)
{
  ATOMO			*Atm1, *Atm2, *InizRes;
  char			File[256];
  FILE			*FH;
  float			Dist, TmpA, TmpB;
  float			*A, *Ai, *Aj, *B, *Bi, *Bj, *Ptr;
  QCHAR			Type;
  LONG			PrecResName, PrecResSeq;
  ULONG			k, j, i;

  struct Ener {
    ATOMO		*InizRes;	/* Pointer to first residue atom */
    float		C;		/* Coulombic energy              */
    float		NB;		/* Lennard-Johnes energy         */
  } *Ene, *AttEne, TmpEne;

  BOOL			NbChk       = FALSE;
  BOOL			Ret         = TRUE;
  float                 CutOff      = Quad(Prefs.ENCUTOFF);
  float			Diel        = Prefs.ENDIEL;
  float			TotalQ	    = 0;
  float			TotalNB     = 0;
  ULONG			Count       = 0;
  ULONG			NAtmRes     = 0;


  if (A = (float *)Alloca(TotAtm * sizeof(float))) {
    if (B = (float *)Alloca(TotAtm * sizeof(float))) {
      PrintProg(MSG_ENER_ASSPROG);
      GetPrgPath(File, TRUE);
      strcat(File, CVFF_PAR);

      /**** Template loader ****/

      if (FH = (FILE *)fopen(File, "r")) {
        if (fgets(File, 255, FH)) {
          if ((*File != ';') && (!strncasecmp(File, "#FF_Param CVFF", 14))) {
            while(fgets(File, 255, FH)) {
              if (!strncasecmp(File, "#NonBond", 8)) {
                NbChk = TRUE;
                while((fgets(File, 255, FH)) && (Count != TotAtm)) {
                  if ((*File != ';') && (*File != '\n')) {
                    Type.L = 0;
                    sscanf(File, "%s %f %f", Type.C, &TmpA, &TmpB);
                    Ai = A;
                    Bi = B;
                    for(Atm1 = InizAtm; Atm1; Atm1 = Atm1 -> Ptr) {
                      if (Atm1 -> Pot.L == Type.L) {
                        ++Count;
                        *Ai = TmpA;
                        *Bi = TmpB;
                      }
                      ++Ai;
                      ++Bi;
                    }
                  }
                }
                if (TotAtm == Count) {
                  if (NAtmRes = FindResPtr(InizAtm, &InizRes, NomeRes, NRes)) {
                    Count = CountRes(InizAtm) - 1;
                    if (Count > 0) {
                      if (Ene = (struct Ener *)Alloca(sizeof(struct Ener) * Count)) {
                        PrintProg(MSG_ENER_CALCPROG);
                        Ai          = A;
                        AttEne      = Ene;
                        Bi          = B;
                        PrecResName = (LONG)NULL;
                        PrecResSeq  = (LONG)NULL;
                        for(Atm1 = InizAtm; Atm1; Atm1 = Atm1 -> Ptr) {
                          if (Atm1 == InizRes) {
                            for(k = 0; k < NAtmRes; ++k) {
                              Atm1 = Atm1 -> Ptr;
                              ++Ai;
                              ++Bi;
                            }
                            if (!Atm1) break;
                          }
                          if ((Atm1 -> ResName.L != PrecResName) || (Atm1 -> ResSeq.L != PrecResSeq)) {
                            if (PrecResName) ++AttEne;
                            AttEne -> InizRes = Atm1;
                            PrecResName = Atm1 -> ResName.L;
                            PrecResSeq  = Atm1 -> ResSeq.L;
                          }

                          Atm2 = InizRes;
                          Aj   = &A[InizRes -> Num - 1];
                          Bj   = &B[InizRes -> Num - 1];
                          for(k = 0; k < NAtmRes; ++k) {

                            /**** Square radius ****/

                            Dist = Quad(Atm1 -> x - Atm2 -> x) +
                                   Quad(Atm1 -> y - Atm2 -> y) +
                                   Quad(Atm1 -> z - Atm2 -> z);

                            if (Dist <= CutOff) {

                              /**** Non-bond term ****/

                              TmpA  = Cube(Dist);
                              TmpB  = (SQR((*Ai) * (*Aj)) / Quad(TmpA)) -
                                      (SQR((*Bi) * (*Bj)) / TmpA);
                              AttEne -> NB += TmpB;
                              TotalNB      += TmpB;

                              /**** Coulombic term ****/

                              TmpB         = CONV_COUL * (Atm1 -> Charge * Atm2 -> Charge) /
                                             (Diel * SQR(Dist));
                              AttEne -> C += TmpB;
                              TotalQ      += TmpB;
                            }

                            Atm2 = Atm2 -> Ptr;
                            ++Aj;
                            ++Bj;
                          }
                          ++Ai;
                          ++Bi;
                        }

                        /**** Sorting energies ****/

                        PrintProg(MSG_ENER_SORTPROG);
                        for(k = 0; k < (Count - 1); ++k) {
                          i    = k;
                          TmpA = Ene[k].NB + Ene[k].C;
                          for(j = k; j < Count; ++j) {
                            TmpB = Ene[j].NB + Ene[j].C;
                            if (TmpB < TmpA) {
                              TmpA = TmpB;
                              i = j;
                            }
                          }
                          memcpy(&TmpEne, &Ene[k], sizeof(struct Ener));
                          memcpy(&Ene[k], &Ene[i], sizeof(struct Ener));
                          memcpy(&Ene[i], &TmpEne, sizeof(struct Ener));
                        }

                        /**** Print results ****/

                        if (fprintf(OUT, GetStr(MSG_ENER_RESULTS),
                                InizRes -> ResName.C, InizRes -> ResSeq.C, NAtmRes,
                                Diel, TotalNB + TotalQ, TotalNB, TotalQ) > 0) {
                          AttEne = Ene;
                          TotalNB += TotalQ;
                          for(k = 0; (Ret) && (k < Count); ++k) {
                            Atm1 = AttEne -> InizRes;
                            TmpA = AttEne -> NB + AttEne -> C;
                            TmpB = (TmpA / TotalNB) * (float)100.0;
                            if (TmpB < Prefs.ENFILTER) break;
                            if (fprintf(OUT, " %-4.4s  %4.4s   %8.4f   %8.4f   %8.4f  %6.2f\n",
                                             Atm1 -> ResName.C, Atm1 -> ResSeq.C, AttEne -> NB,
                                             AttEne -> C, TmpA, TmpB) <= 0)
                              Ret = PrintDosErr();
                            ++AttEne;
                          }
                          if (fprintf(OUT, "\n\n") <= 0) Ret = PrintDosErr();
                        } else Ret = PrintDosErr();
                        FREE(Ene);
                      } else Ret = FALSE;
                    } else Ret = CatErr(MSG_ERR_ENER_MORERES);
                  } else Ret = CatErr(MSG_ERR_ENER_RESNOTFOUND);
                } else {
                  Ptr = A;
                  for(Atm1 = InizAtm; Atm1; Atm1 = Atm1 -> Ptr) {
                    if (*Ptr == 0.0) {
                      sprintf(File, GetStr(MSG_ERR_ENER_UNKATMTYPE), Atm1 -> Pot.C);
                      PrintAtmErr(File, Atm1);
                      Ret = FALSE;
                    }
                    ++Ptr;
                  }
                }
                break;
              }
            }
            if (!NbChk) Ret = CatErr(MSG_ERR_ENER_NONBOND);
          }
        }
        fclose(FH);
      } else Ret = CatErr(MSG_ERR_ENER_FFNOTFOUND, CVFF_PAR);
      FREE(B);
    } else Ret = FALSE;
    FREE(A);
  } else Ret = FALSE;

  return Ret;
}

