/*

             **************************************************
             ****             VEGA - Locale V1.0           ****
             ****           Localization routines          ****
             **** Copyright 1996-2000, Alessandro Pedretti ****
             **************************************************

The following routines use the AmigaOS Locale Technology to translate the
strings in any language using a specific catalog file in standard IFF format.
Each locale file must be stored in a sub-directory with the lower-case language
name of the Locale directory and must have VEGA.catalog file name.
To create a new catalog, you must use FlexCat compiler with the catalog
description file (.cd) and the catalog translation file (.ct). Please read the
specific documentation in FlexCat directory and in VEGA manual.
The AmigaOS Locale Technology is copyrighted by Amiga Inc.

*/

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "globdef.h"
#include "globvar.h"
#include "iff.h"
#define  VEGA_ARRAY
#include "globstr.h"

#ifdef AMIGA
#  include <exec/libraries.h>
#  include <libraries/locale.h>

    extern struct Library  *LocaleBase;
#else
#  ifdef WIN32
#    define  LOCALE_DIR   "Catalogs\\"
#  else
#    define  LOCALE_DIR   "Catalogs/"
#  endif

typedef struct {
  ULONG         Ver, Rel;
  char          Name[32];
  char          Date[14];
} AMIGAVER;

static void Version(char *, AMIGAVER *);
#endif

#ifdef WIN32
#define  ERROR_PROMPT "\n%s %s.\n\n"
#else
#define  ERROR_PROMPT "\n\033[33m%s\033[0m %s.\n\n"
#endif

/**** Local variables ****/

static ULONG      CatLen;


/**** Shows an error ****/

BOOL CatErr(ULONG Err, ...)
{
  char          Buf[128];
  va_list       vl;

  sprintf(Buf, ERROR_PROMPT, GetStr(MSG_ERR_LOCALE_ERROR), GetStr(Err));
  va_start(vl, Err);
  vfprintf(stderr, Buf, vl);
  va_end(vl);

  return FALSE;
}


/**** fprintf replacement with string localization ****/

int CatPrintf(FILE *FH, ULONG Str, ...)
{
  int           Ret;
  va_list       vl;

  va_start(vl, Str);
  Ret = vfprintf(FH, GetStr(Str), vl);
  va_end(vl);

  return Ret;
}


/**** Translate a catalog string                 ****/
/**** compatible with AmigaOS GetCatalogString() ****/

#ifndef AMIGA
char *GetCatalogStr(void *Cat, register ULONG Id, char *BuiltIn)
{
  register ULONG        *Ptr, Len;
  register ULONG        CatEnd = (ULONG)Cat + CatLen;

  Ptr = (ULONG *)Cat;
  while((ULONG)Ptr < CatEnd) {
    if (Id == *Ptr) {
      BuiltIn = (char *)Ptr + 8;
      break;
    }
    Len = Ptr[1];
    if (Len & 3) Len += 4 - (Len & 3);
    (char *)Ptr += Len + 8;
  }

  return BuiltIn;
}
#endif


/**** Return a string and if possible, translate it ****/

char *GetStr(ULONG Id)
{
  char                  *BuiltIn = "Unknown localized string";
  register ULONG        k        = sizeof(VEGA_Str) / sizeof(LOCARRAY);
  register LOCARRAY     *Ptr     = (LOCARRAY *)VEGA_Str;

  while(k--) {
    if (Ptr -> Id == Id) {
      BuiltIn = Ptr -> Str;
    }
    ++Ptr;
  }

  /**** String localization ****/

  if (Catalog)
    return((char *)GetCatalogStr((void *)Catalog, Id, BuiltIn));

  return BuiltIn;
}


/****       Open a catalog file      ****/
/**** (only for non-AmigaOS systems) ****/

#ifndef AMIGA
void *OpenCatalog(char *CatalogName, char *Language, ULONG MinVer)
{
  AMIGAVER      Ver;
  char          CatalogPath[512];
  FILE          *FH;
  IFFHDR        Hdr;
#ifdef LITTLE_ENDIAN
  ULONG         *Ptr, Len;
#endif

  BOOL          Ret  = TRUE;
  const char    *Err = "Corrupted catalog file";
  UWORD         Chk  = 0;
  void          *Cat = NULL;

  if (strcmp(Language, "english")) {

    /**** Set the path and file name ****/

    GetPrgPath(CatalogPath, FALSE);
    strcat(CatalogPath, LOCALE_DIR);
    if (!strcmp(Language, "auto")) {
#ifdef WIN32
      if ((Language = GetSysLangStr()) == NULL)
        return NULL;
#else
      PrintWarn("The automatic language detection is not supported\n" \
                "from your operating system. Please set the key LANGUAGE\n" \
                "of the prefs file with an appropriate language");
      return NULL;
#endif
    }
    strcat(CatalogPath, Language);
    strcat(CatalogPath, C_DIRSEP);
    strcat(CatalogPath, CatalogName);

    /**** Read the IFF file ****/

    if (FH = OpenIff(CatalogPath, IFF_CTLG)) {
      while((Ret) && (ReadIffHdr(FH, &Hdr, IFF_READ_NODOSERR))) {
        switch(Hdr.ID) {

          case IFF_CTLG_FVER:  /* Version chunk */
            if ((Hdr.Size <= sizeof(CatalogPath)) &&
                (fread(CatalogPath, Hdr.Size, 1, FH) == 1)) {
              Version(CatalogPath, &Ver);
              if (!strcasecmp(CatalogName, Ver.Name)) {
                if (Ver.Ver < MinVer)
                  Ret = Errore("%s" C_DIRSEP "%s %d.%d\n"  \
                               "The version %d.0 or better is required",
                               CatalogName, Language, Ver.Ver, Ver.Rel, MinVer);
                else ++Chk;
              } else Ret = Errore("%s" C_DIRSEP "%s\n" \
                                  "Catalog of another application",
                                  Language, CatalogName);
            } else Ret = Errore(Err);
            break;

          case IFF_CTLG_LANG:  /* Language chunk */
            if ((Hdr.Size <= sizeof(CatalogPath)) &&
                (fread(CatalogPath, Hdr.Size, 1, FH) == 1)) {
              if (strcasecmp(CatalogPath, Language))
                Ret = Errore("%s" C_DIRSEP "%s\n" \
                             "has a wrong language(%s)",
                             Language, CatalogName, CatalogPath);
              else ++Chk;
            } else Ret = Errore(Err);
            break;

          case IFF_CTLG_STRS:  /* String chunk */
            if (Cat = (void *)Alloca(Hdr.Size)) {
              if (fread(Cat, Hdr.Size, 1, FH) == 1) {
                CatLen = Hdr.Size;

                /**** Set the endian (if needed) ****/

#ifdef LITTLE_ENDIAN
                Ptr = (ULONG *)Cat;
                while((ULONG)Ptr < ((ULONG)Cat + Hdr.Size)) {
                  Swap(Ptr);
                  Swap(Ptr + 1);
                  Len = Ptr[1];
                  if (Len & 3) Len += 4 - (Len & 3);
                  (char *)Ptr += 8;

                  /**** Convert the character set ****/

#ifdef WIN32
                  Iso2Dos((char *)Ptr);
#endif
                   (char *)Ptr += Len;
                }
#endif
                ++Chk;
              } else {
                Ret = Errore(Err);
                FREE(Cat);
                Cat = NULL;
              }
            }
            break;

          default:            /* Skip the unknown chunk */
            fseek(FH, Hdr.Size, SEEK_CUR);
        }
      }
      if ((Chk != 3) && (Cat)) {
        FREE(Cat);
        Cat = NULL;
      }
      fclose(FH);
    }
  } /* The english language is built-in and don't need to load */

  return Cat;
}


/**** Decode a standard AmigaOS version string ****/

static void Version(register char *Str, AMIGAVER *Ver)
{
  *Ver -> Name = NULL;
  *Ver -> Date = NULL;
  Ver -> Ver   = 0;
  Ver -> Rel   = 0;

  if (!strncmp(Str, "$VER", 4)) {
    Str += 5;
    while((*Str) && (*Str == ' ')) ++Str;
    sscanf(Str, "%31s %d.%d (%13s ", Ver -> Name, &Ver -> Ver,
           &Ver -> Rel, Ver -> Date);
    Ver -> Date[strlen(Ver -> Date) - 1] = 0;
  }
}
#endif


