/*

             **************************************************
             ****     VEGA - un/packing utilities V1.0     ****
             **** Copyright 1996-2000, Alessandro Pedretti ****
             **************************************************


Universal data de/compressor engine for VEGA.

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "globdef.h"
#include "globvar.h"


/**** Local prototypes ****/

BOOL BZ2_DepackFile(char *, char *);
BOOL GZ_DepackFile(char *, char *);
BOOL PP_DepackFile(char *, char *);
BOOL Z_DepackFile(char *, char *);

static FILE *MyOpen(char *, char *, BOOL);
static void GetTmpFileName(char *Str);


/**** Packer headres in little-endian format ****/

#define  PK_HDR_BZIP2               0x425A6830    /* Bzip2 1-9         */
#define  PK_HDR_COMPRESS            0x1F9D0000    /* Unix Z compress   */
#define  PK_HDR_GZIP                0x1F8B0000    /* GZip magic number */
#define  PK_HDR_POWERPACKER         0x50503230    /* PP20              */
#define  PK_HDR_XPK                 0x58504B46    /* XPKF              */


/**** Local definitions ****/

typedef struct __PkFileInfo {
  struct __PkFileInfo   *Ptr;           /* Pointer to next list element      */
  FILE                  *FH;            /* File handle                       */
  char                  FileName[512];  /* File name with path               */
  BOOL                  NoRemove;       /* Don't remove the file when closed */
} PKFILEINFO;


/**** Local variables ****/

PKFILEINFO      *InizPkInfo = NULL;



/**** Get a temporary file name with path ****/

void GetTmpFileName(char *Str)
{
  char          *Ptr, Tmp[32];

  tmpnam(Tmp);

#if defined(WIN32)
  if (((Ptr = getenv("TEMP")) == NULL) &&
      ((Ptr = getenv("TMP")) == NULL))
     Ptr = "";
  strcpy(Str, Ptr);
  Ptr = Str + strlen(Str) - 1;
  if (*Ptr != '\\') strcat(Str, "\\");
#else
  *Str = 0;
#endif

  strcat(Str, Tmp);
}


/**** Close a temporary file and free all resources ****/

void MyClose(FILE *FH)
{
  PKFILEINFO            *PkInfo, *PrecPkInfo;

  if (FH) {

    /**** Find the entry into the list ****/

    PkInfo     = InizPkInfo;
    while((PkInfo) && (PkInfo -> FH != FH)) {
      PrecPkInfo = PkInfo;
      PkInfo     = PkInfo -> Ptr;
    }
    fclose(FH);
    if (PkInfo) {
      if ((!PkInfo -> NoRemove) && (remove(PkInfo -> FileName)))
        PrintDosErr();                                /* Remove the file         */
      if (PkInfo == InizPkInfo) InizPkInfo = NULL;    /* Remake the list binding */
      else PrecPkInfo -> Ptr = PkInfo -> Ptr;
      FREE(PkInfo);                                   /* Free the memory         */
    }
  }
}


/**** Close all opened file and remove the temporary files ****/

void MyCloseAll(void)
{
  while(InizPkInfo)
    MyClose(InizPkInfo -> FH);
}


/**** Open a temporary file and store all infos ****/

FILE *MyOpen(char *FileName, char *Mode, BOOL NoRemove)
{
  FILE          *FH;
  PKFILEINFO    *New, *PkInfo;

  if (FH = fopen(FileName, Mode)) {
    if (New = (PKFILEINFO *)Alloca(sizeof(PKFILEINFO))) {
      strcpy(New -> FileName, FileName);
      New -> FH       = FH;
      New -> NoRemove = NoRemove;
      if (InizPkInfo) {

        /**** Add the entry into the list ****/

        for(PkInfo = InizPkInfo; PkInfo -> Ptr; PkInfo = PkInfo -> Ptr);
        PkInfo -> Ptr = New;
      } else InizPkInfo = New;
    } else {
      fclose(FH);
      FH = NULL;
    }
  } else PrintDosErr();

  return FH;
}


/**** Open & decrunch a packed file ****/

FILE *PkOpen(char *FileName, char *Mode, BOOL Verbose)
{
  char          TmpFileName[256];
  LONG          Hdr;

  BOOL          Ret   = FALSE;
  FILE          *FH   = NULL;

  /**** Set the name of the temporary file ****/

  GetTmpFileName(TmpFileName);

  /**** Read a packed file ****/

  if (*Mode == 'r') {
    if (FH = fopen(FileName, "rb")) {
      if (fread(&Hdr, sizeof(LONG), 1, FH) == 1) {
        fclose(FH);
        FH = NULL;

#ifdef LITTLE_ENDIAN
        Swap(&Hdr);
#endif

        /**** Packer recognition ****/

        switch(Hdr) {
        case PK_HDR_POWERPACKER:   /* Amiga PowerPacker */
          Ret = PP_DepackFile(FileName, TmpFileName);
          break;

        case PK_HDR_XPK:           /* Amiga/Unix xPK master */
          Errore("Xpk compression not yet implemented");
          return NULL;

        default:

          switch (Hdr & 0xFFFF0000) {
          case PK_HDR_GZIP:        /* GZip */
            Ret = GZ_DepackFile(FileName, TmpFileName);
            break;

          case PK_HDR_COMPRESS:    /* Unix compress */
            Ret = Z_DepackFile(FileName, TmpFileName);
            break;

          default:
            if ((Hdr & 0xFFFFFFF0) == PK_HDR_BZIP2) /* BZip2 */
              Ret = BZ2_DepackFile(FileName, TmpFileName);
          } /* End of switch */
        } /* End of switch */

        /**** Open the file ****/

        if (Ret) FH = MyOpen(TmpFileName, Mode, FALSE);
        else FH = MyOpen(FileName, Mode, TRUE);
      } else if (Verbose) PrintDosErr();
    } else if (Verbose) PrintDosErr();
  }

  return FH;
}


/**** Re-open a file without depacking ****/
/**** Useful to chenge the stream mode ****/

FILE *PkReOpen(FILE *FH, char *Mode)
{
  PKFILEINFO            *PkInfo, *PrecPkInfo;

  if (FH) {
    PkInfo     = InizPkInfo;
    while((PkInfo) && (PkInfo -> FH != FH)) {
      PrecPkInfo = PkInfo;
      PkInfo     = PkInfo -> Ptr;
    }
    if (PkInfo) {
      fclose(FH);
      if ((PkInfo -> FH = fopen(PkInfo -> FileName, Mode)) == NULL) {
        PrintDosErr();
        if (!PkInfo -> NoRemove) remove(PkInfo -> FileName); /* Remove the file         */
        if (PkInfo == InizPkInfo) InizPkInfo = NULL;         /* Remake the list binding */
        else PrecPkInfo -> Ptr = PkInfo -> Ptr;
        FREE(PkInfo);                                        /* Free the memory         */
      }
    }
  }

  return PkInfo -> FH;
}





